import aws_cdk.aws_apigateway as apigateway
from constructs import Construct


class ApiGatewayFactory:
    def __init__(self, scope: Construct, name: str):
        self.api = apigateway.RestApi(scope, name)

    def add_lambda(self, handler):
        self.api.root.add_resource()


from aws_cdk.aws_ec2 import SecurityGroup, Port
from constructs import Construct, IConstruct
from infraflow.cdk.sg import tier_maps
from infraflow.cdk.sg.ports import all_tcp


class SecurityGroupNeeds:
    def __init__(
            self,
            needs_outbound_internet=None,
            needs_inbound_internet=None
    ):
        self.needs_inbound_internet = needs_inbound_internet
        self.needs_outbound_internet = needs_outbound_internet


class SecurityGroupTarget:
    def __init__(
            self,
            scope: Construct, id: str, *args,
            cdk_type: type = None,
            infraflow_pattern: object = None,
    ):
        self.infraflow_pattern = infraflow_pattern
        self.cdk_type = cdk_type
        self.id = id
        self.scope = scope


class SecurityGroupStructurePattern:
    def __init__(self, stack):
        self.stack = stack

    def get_group(
            self,
            target: SecurityGroupTarget,
            tier: str = None,
            needs: SecurityGroupNeeds = SecurityGroupNeeds()

    ) -> SecurityGroup:
        raise NotImplemented()


class SingleSecurityGroup(SecurityGroupStructurePattern):
    def __init__(self, stack):
        super().__init__(stack)

    def get_group(
            self,
            target: SecurityGroupTarget,
            tier: str = None,
            needs: SecurityGroupNeeds = SecurityGroupNeeds()

    ) -> SecurityGroup:
        return self.stack.default_sg


class SecurityGroupPerResource(SecurityGroupStructurePattern):
    def __init__(self, stack):
        super().__init__(stack)

    def get_group(
            self,
            target: SecurityGroupTarget,
            tier: str = None,
            needs: SecurityGroupNeeds = SecurityGroupNeeds()

    ) -> SecurityGroup:
        return SecurityGroup(scope=target.scope, id=target.id+"Sg")


class Tiered(SecurityGroupStructurePattern):
    def __init__(self, stack, db_ports: list[Port], app_web_ports: list[Port] = [all_tcp], allow_all_within_app_group=True, allow_all_within_group: bool = False):
        super().__init__(stack)
        self.dbs = SecurityGroup(self.stack, 'DbSg', vpc=self.stack.env.vpc)
        self.app = SecurityGroup(self.stack, 'AppSg', vpc=self.stack.env.vpc)
        self.web = SecurityGroup(self.stack, 'WebSg', vpc=self.stack.env.vpc)
        for port in app_web_ports:
            self.web.connections.allow_to(self.app, port)
        for port in db_ports:
            self.app.connections.allow_to(self.dbs, port)

        if allow_all_within_group or allow_all_within_app_group:
            self.app.connections.allow_to(self.app, Port.all_tcp())

        if allow_all_within_group:
            self.dbs.connections.allow_to(self.dbs, Port.all_tcp())
            self.web.connections.allow_to(self.web, Port.all_tcp())

    def get_group(
            self,
            target: SecurityGroupTarget,
            tier: str = None,
            needs: SecurityGroupNeeds = SecurityGroupNeeds()

    ) -> SecurityGroup:
        if tier == 'app':
            return self.app
        elif tier == 'db':
            return self.dbs
        elif tier == 'web':
            return self.web
        elif target.cdk_type in tier_maps.app_services:
            return self.app
        elif target.cdk_type in tier_maps.web_services:
            return self.web
        elif target.cdk_type in tier_maps.db_services:
            return self.dbs
        else:
            return self.stack.default_sg




from aws_cdk import IResource
import aws_cdk.aws_iam as iam
from aws_cdk.aws_iam import Policy, PolicyProps
from constructs import Construct

from infraflow.cdk.arns import arn_for_resource, service_key_for_resource


class IamResource:
    def __init__(self, resource: IResource = None, service=None, pattern: str = None, all: bool = False):
        if (int(bool(resource)) + int(bool(pattern)) + int(all)) > 1:
            raise ValueError('Must only provide one parameter for: resource, pattern, or all')
        self.all = all
        self.resource = resource

        if pattern:
            if service:
                self.resource_string = f"{service}:pattern"
            else:
                self.resource_string = pattern
        elif all:
            self.resource_string = f'{service}:*'
        elif resource:
            self.resource_string = arn_for_resource(resource)

        if resource:
            self.service = service_key_for_resource(resource)
        else:
            self.service = service


class IamAction:
    def __init__(self, service: str, all: bool = False, pattern: str = None):
        """
        :param service: What AWS service this action is for
        :param all: whether to allow all actions for this service, can't be true/set if pattern is set
        :param pattern: the pattern to use for this service, can't be set if all is true
        """
        if all and pattern:
            raise ValueError('Must only provide one parameter for: pattern, or all')
        self.all = all
        if pattern:
            self.pattern = pattern
        elif all:
            self.pattern = '*'
        self.service = service
        self.action_string = f"{self.service}:{self.pattern}"


class IamStatement:
    def __init__(self,
                 effect: iam.Effect = iam.Effect.ALLOW,
                 resources: list[IamResource] = [],
                 actions: list[IamAction] = []):
        self.effect = effect
        self.actions = actions
        self.resources = resources

    def allow(self):
        self.effect = iam.Effect.ALLOW
        return self

    def deny(self):
        self.effect = iam.Effect.DENY
        return self

    def action(self, action: IamAction):
        self.actions.append(action)

    def for_actions(self, *actions: IamAction):
        self.actions.extend(actions)

    def to_cdk(self):
        return iam.PolicyStatement(
            actions=[a.action_string for a in self.actions],
            resources=[r.resource_string for r in self.resources],
            effect=self.effect
        )


class PolicyBuilder:
    def __init__(self):
        self.statements: list[IamStatement] = []

    def allow_resource(self, resource: IResource):
        statement = IamStatement(
            effect=iam.Effect.ALLOW,
            resources=[IamResource(resource)]
        )
        self.statements.append(statement)
        return statement

    def allow_resources(self, resources: list[IResource]):
        statement = IamStatement(
            effect=iam.Effect.ALLOW,
            resources=[IamResource(resource) for resource in resources]
        )
        self.statements.append(statement)
        return statement

    def allow_on_all(self, service: str):
        statement = IamStatement(
            effect=iam.Effect.ALLOW,
            resources=[IamResource(service=service, all=True)]
        )
        self.statements.append(statement)
        return statement

    def allow_pattern(self, service: str, pattern: str):
        statement = IamStatement(
            effect=iam.Effect.ALLOW,
            resources=[IamResource(service=service, pattern=pattern)]
        )
        self.statements.append(statement)
        return statement

    def deny_resource(self, resource: IResource):
        statement = IamStatement(
            effect=iam.Effect.DENY,
            resources=[IamResource(resource)]
        )
        self.statements.append(statement)
        return statement

    def deny_resources(self, resources: list[IResource]):
        statement = IamStatement(
            effect=iam.Effect.DENY,
            resources=[IamResource(resource) for resource in resources]
        )
        self.statements.append(statement)
        return statement

    def deny_on_all(self, service: str):
        statement = IamStatement(
            effect=iam.Effect.DENY,
            resources=[IamResource(service=service, all=True)]
        )
        self.statements.append(statement)
        return statement

    def deny_pattern(self, service: str, pattern: str):
        statement = IamStatement(
            effect=iam.Effect.DENY,
            resources=[IamResource(service=service, pattern=pattern)]
        )
        self.statements.append(statement)
        return statement

    def build_cdk_policy(self, parent: Construct, id: str) -> Policy:
        return Policy(parent, id, PolicyProps(statements=[s.to_cdk() for s in self.statements]))



from typing import Optional, Any

from aws_cdk import App, aws_sns, IResource
import aws_cdk.aws_lambda as aws_lambda
import aws_cdk.aws_apigateway as apigateway
import aws_cdk.aws_events as aws_events
from aws_cdk.aws_ec2 import InterfaceVpcEndpointAwsService
from aws_cdk.aws_iam import IRole, Role, ServicePrincipal
from aws_cdk.aws_stepfunctions import IStateMachine

from infraflow.cdk import ServiceStageStack, EnvConfig
from infraflow.cdk.core.environment import DEFAULT_INTERFACE_SERVICES
from infraflow.cdk.docker import EcsCluster
from infraflow.cdk.events import EventBridgeEvents, InfraflowEventBus, SnsEvents
from infraflow.cdk.iam import PolicyBuilder, IamAction
from infraflow.cdk.lambdas import LambdaContext
from infraflow.cdk.sg.patterns import Tiered
from infraflow.cdk.sg.ports import port_for


class StandardServiceStage(ServiceStageStack):
    # noinspection PyDefaultArgument
    def __init__(self,
                 app: App,
                 service_name: str,
                 stage_name: str,
                 env: EnvConfig,
                 src_path: str = "src",
                 db_ports=[],
                 python_version: aws_lambda.Runtime = aws_lambda.Runtime.PYTHON_3_9,
                 vpc_endpoint_services: list[InterfaceVpcEndpointAwsService] = DEFAULT_INTERFACE_SERVICES,
                 **kwargs):
        super().__init__(app, service_name, stage_name, env, **kwargs)
        self.db_ports = db_ports
        self.vpc_endpoint_services = vpc_endpoint_services
        self._app_role = None
        self._event_publish_policy = None
        self.python_version = python_version
        self.src_path = src_path
        self._lambda_context: Optional[LambdaContext] = None
        self._api: Optional[apigateway.IRestApi] = None
        self._event_bridge_bus_cdk: Optional[aws_events.IEventBus] = None
        self._events: Optional[EventBridgeEvents] = None
        self._ecs_cluster: Optional[EcsCluster] = None
        self._sns_bus_cdk: Optional[aws_sns.Topic] = None
        self.security_groups: Tiered = Tiered(self, db_ports=db_ports)

    @property
    def api(self) -> apigateway.IRestApi:
        self.use_api()
        return self._api

    def use_api(self):
        self._api = self._api or apigateway.RestApi(self, "API")

    @property
    def events(self) -> InfraflowEventBus:
        self.use_events()
        return self._events

    def use_events(self):
        if not self._event_bridge_bus_cdk and not self._sns_bus_cdk:
            self.use_event_bridge()

        self._events = self._events or (
            EventBridgeEvents(self, 'Events', self._event_bridge_bus_cdk) if self._event_bridge_bus_cdk else
            SnsEvents(self, 'Events', self._sns_bus_ckd) if self._sns_bus_ckd else
            None
        )

        if self._event_bridge_bus_cdk and self._app_role:
            self._event_bridge_bus_cdk.grant_put_events_to(self._app_role)

    @property
    def event_bridge_bus(self) -> aws_events.IEventBus:
        self.use_event_bridge()
        return self._event_bridge_bus_cdk

    def use_event_bridge(self):
        self._event_bridge_bus_cdk = self._event_bridge_bus_cdk or aws_events.EventBus(self, "Bus")

    @property
    def sns_bus(self) -> aws_sns.ITopic:
        self.use_sns()
        return self._sns_bus_cdk

    def use_sns(self):
        self._sns_bus_cdk = self._sns_bus_cdk or aws_sns.Topic(self, "SnsTopic")

    @property
    def lambda_context(self) -> LambdaContext:
        self.use_lambda()
        return self._lambda_context

    def use_lambda(self):
        self._lambda_context = self._lambda_context or LambdaContext(
            self,
            path=self.src_path,
            role=self.app_role,
            runtime=self.python_version
        )

    @property
    def ecs_cluster(self) -> EcsCluster:
        self.use_ecs()
        return self._ecs_cluster

    def use_ecs(self):
        self._ecs_cluster = self._ecs_cluster or EcsCluster(self, "EcsCluster")

    def use_monolith_lambda_api(self, handler):
        self._api = apigateway.LambdaRestApi(
            self, "API", handler=self.lambda_context.api_function(handler, "ApiHandler")
        )

    def use_monolith_step_function_api(self, state_machine: IStateMachine):
        self._api = apigateway.StepFunctionsRestApi(
            self, "API", state_machine=state_machine
        )

    def lambda_api_resource_handler(self, handler, name=None, suffix=None):
        return apigateway.LambdaIntegration(
            handler=self.lambda_context.api_function(handler, name=name, suffix=suffix)
        )

    # noinspection PyMethodMayBeStatic
    def step_functions_api_resource_handler(self, state_machine: IStateMachine):
        integration = apigateway.StepFunctionsIntegration()
        integration.start_execution(state_machine)
        return integration

    @property
    def app_group(self):
        return self.security_groups.app

    @property
    def dbs_group(self):
        return self.security_groups.dbs

    @property
    def web_group(self):
        return self.security_groups.web

    def use_external_resource(self, resource: IResource, name: str, actions: list[IamAction] = None):
        db: Any = resource
        self.app_group.connections.allow_to(resource, port_range=port_for(db))
        if actions and len(actions):
            policy = PolicyBuilder()
            policy.allow_resource(resource).for_actions(actions)
            self.app_role.attach_inline_policy(policy.build_cdk_policy(self, f"{name}Policy"))

    @property
    def app_role(self) -> IRole:
        self._app_role = self._app_role or Role(self, 'Role', assumed_by=ServicePrincipal("lambda.amazonaws.com"))
        if self._event_bridge_bus_cdk:
            self._event_bridge_bus_cdk.grant_put_events_to(self._app_role)
        return self._app_role

    def setup_endpoint_access(self):
        interfaces = self.env.service_endpoints(self.vpc_endpoint_services)
        for x in interfaces:
            self.security_groups.app.connections.allow_to(x)



from typing import Union

from aws_cdk import IResource
from aws_cdk.aws_ec2 import Port, Protocol
from aws_cdk import aws_rds as rds
from aws_cdk import aws_elasticache as cache
from aws_cdk import aws_elasticsearch as search

postgres_port = Port.tcp(5432)
mysql_port = Port.tcp(3306)
redis_port = Port.tcp(6379)
opensearch_port = Port.tcp(9200)
https_port = Port.tcp(443)
http_port = Port.tcp(80)
http_alt_port = Port.tcp(8080)
all_tcp = Port.all_tcp()


def port_for(resource: Union[rds.DatabaseInstance, search.IDomain]):
    if isinstance(resource, rds.DatabaseInstance):
        return Port.tcp(resource.instance_endpoint.port)
    elif isinstance(resource, search.Domain):
        return opensearch_port






import datetime
from typing import Union

from aws_cdk import aws_lambda, Duration
from aws_cdk import aws_sqs
from aws_cdk import aws_lambda_event_sources as sources
from aws_cdk.aws_ec2 import SubnetType, SubnetSelection
from aws_cdk.aws_events import Rule, Schedule
from aws_cdk.aws_events_targets import LambdaFunction
from aws_cdk.aws_iam import IRole
from aws_cdk.aws_lambda import IEventSource
from constructs import Construct, IConstruct

from infraflow.cdk.core.service_stage import ServiceStageStack
from infraflow.cdk.iam import PolicyBuilder, action_groups
from infraflow.cdk.sg.patterns import SecurityGroupTarget
from infraflow.util import caps_camel

registry = []


class LambdaContext:
    def __init__(self,
                 stage: ServiceStageStack,
                 path: str,
                 role: IRole = None,
                 runtime: aws_lambda.Runtime = aws_lambda.Runtime.PYTHON_3_9,
                 vpc: bool = True,
                 subnet_type: SubnetType = SubnetType.PRIVATE_WITH_EGRESS,
                 tracing=aws_lambda.Tracing.ACTIVE
                 ):
        self.role = role
        self.runtime = runtime
        self.path = path
        self.stage = stage
        self.default_sg = stage.default_sg

    # def to_cdk(self):
    #     return dict(
    #         scope=self.scope,
    #         code=aws_lambda.Code.from_asset(self.path),
    #         runtime=self.runtime
    #     )

    def function(self, handler: str, name=None, suffix=None, scope_override: IConstruct = None):
        constructed_name = self.construct_name(handler, name, suffix)
        func = aws_lambda.Function(
                id=constructed_name,
                handler=handler,
                scope=scope_override or self.stage,
                code=aws_lambda.Code.from_asset(self.path),
                runtime=self.runtime,
                security_groups=[
                    self.stage.security_groups.get_group(target=SecurityGroupTarget(
                        scope_override or self.stage,
                        id=constructed_name,
                        cdk_type=aws_lambda.Function,
                        infraflow_pattern=self,
                    ))
                ],
                vpc_subnets=SubnetSelection(subnets=self.stage.env.vpc_subnets()),
                vpc=self.stage.env.vpc,
                tracing=aws_lambda.Tracing.ACTIVE,
                role=self.role
                # function_name=''
        )
        # if self.role:
        #     func.grant_invoke(self.role)
        return func

    def construct_name(self, handler: str, name: str, suffix: str):
        base_name = name or caps_camel(handler.split(".")[1])
        constructed_name = caps_camel(f"{base_name}_{suffix}" if suffix else base_name)
        return constructed_name

    def queued_function(self, handler, name=None, suffix=None, **queue_config):
        constructed_name = self.construct_name(handler, name, suffix)
        queued_function = QueueFunctionConstruct(self.stage, constructed_name)
        queue = aws_sqs.Queue(queued_function, 'Queue', **queue_config)
        func = self.function(handler, 'Function', scope_override=queued_function)
        func.add_event_source(sources.SqsEventSource(queue))
        queued_function.queue = queue
        queued_function.function = func
        return queued_function

    def scheduled_function(self, handler, schedule: Union[Schedule, Duration, datetime.timedelta], name=None, suffix=None):
        schedule = Duration.seconds(schedule.total_seconds()) if isinstance(schedule, datetime.timedelta) else None
        schedule = Schedule.rate(schedule) if isinstance(schedule, Duration) else None
        constructed_name = self.construct_name(handler, name, suffix)
        func = self.function(handler, name, suffix)
        rule = Rule(self.stage, f"{constructed_name}_schedule", schedule=schedule)
        rule.add_target(LambdaFunction(func))
        return func

    def api_function(self, handler, name=None, suffix=None):
        return self.function(handler, name, suffix)

    def function_with_source(self, handler, event_source: IEventSource, name=None, suffix=None):
        func = self.function(handler, name, suffix)
        func.add_event_source(event_source)
        return func


class QueueFunctionConstruct(Construct):
    function: aws_lambda.Function
    queue: aws_sqs.Queue

    def __init__(self, scope: Construct, id: str):
        super().__init__(scope, id)



from aws_cdk import aws_ecs as ecs
from aws_cdk import aws_ecs_patterns as ecs_patterns
from aws_cdk import aws_ecr_assets as assets
from aws_cdk.aws_ec2 import SubnetSelection
from aws_cdk.aws_ecs import Cluster

from infraflow.cdk import ServiceStageStack
from infraflow.cdk.sg.patterns import SecurityGroupTarget


class EcsCluster:
    def __init__(self, scope: ServiceStageStack, cluster_name: str):
        self.scope = scope
        vpc = self.scope.env.vpc

        self.cluster = ecs.Cluster(self.scope, cluster_name, vpc=vpc)

    def service(self,
                name,
                ecr_image: str = None,
                path: str = None,
                command=None,
                count=1,
                cpu=256,
                memory_limit_mib=512
                ):
        if path:
            image_asset = assets.DockerImageAsset(self.scope, f"{name}_image", directory=path)
            image = ecs.ContainerImage.from_docker_image_asset(image_asset)
        else:
            image = None

        return ecs_patterns.ApplicationLoadBalancedFargateService(
            self.scope, f"{name}_service",
            cluster=self.cluster,  # Required
            cpu=cpu,  # Default is 256
            desired_count=count,  # Default is 1
            listener_port=80,
            task_image_options=ecs_patterns.ApplicationLoadBalancedTaskImageOptions(
                image=ecs.ContainerImage.from_registry(ecr_image) if ecr_image else image if path else None,
                container_name=f"{name}_task",
                container_port=80,
                command=command
            ),
            security_groups=[
                self.scope.security_groups.get_group(target=SecurityGroupTarget(
                    self.cluster,
                    id=name,
                    cdk_type=Cluster,
                    infraflow_pattern=self,
                ))
            ],
            task_subnets=SubnetSelection(subnets=self.scope.env.vpc_subnets()),
            memory_limit_mib=memory_limit_mib,  # Default is 512
            public_load_balancer=True
        )  # Default is True




from constructs import Construct

from infraflow.cdk.core.construct import HasEnv, ConstructWithEnv, HasDefaultSg


class ComponentService(Construct, HasEnv, HasDefaultSg):
    def __init__(self, parent:  ConstructWithEnv, id: str, **kwargs):
        super().__init__(parent, id, **kwargs)
        self.env = parent.env
        self.default_sg = parent.default_sg



from aws_cdk import Stack, App
from aws_cdk.aws_ec2 import SecurityGroup

from infraflow.cdk.core.construct import HasEnv, HasDefaultSg
from infraflow.cdk.core.environment import EnvConfig, Env
from infraflow.cdk.sg.patterns import SecurityGroupStructurePattern, SecurityGroupTarget


class ServiceStageStack(Stack, HasEnv, HasDefaultSg):
    def __init__(
            self,
            app: App,
            service_name: str,
            stage_name: str,
            env: EnvConfig,
            **kwargs
    ):
        super().__init__(app, f"{stage_name}-{service_name}", env=env.env, **kwargs)
        self.service_name = service_name
        self.stage_name = stage_name
        self.scope = app
        self.env = Env(stage_name, self, env)
        self.default_sg: SecurityGroup = None
        self.security_groups: SecurityGroupStructurePattern = None

    def get_sg(self, id=None, cdk_type: type = None):
        return self.security_groups.get_group(SecurityGroupTarget(self.scope, id=id, cdk_type=cdk_type))



from typing import TypeVar

from aws_cdk.aws_ec2 import SecurityGroup
from constructs import Construct

from infraflow.cdk.core.environment import IEnv


class HasEnv:
    env: IEnv


class HasDefaultSg:
    default_sg: SecurityGroup


ConstructWithEnv = TypeVar('ConstructWithEnv', Construct, HasEnv, HasDefaultSg)



from aws_cdk import aws_stepfunctions_tasks as tasks
from aws_cdk import aws_stepfunctions as sfn
from aws_cdk import aws_events as events
from aws_cdk.aws_stepfunctions import TaskInput

from infraflow.cdk.environment import ServiceStage


class AwaitBatch:
    def __init__(self, environment: ServiceStage):
        self.environment = environment

        name, data, type, bus = None, None, None, ''
        self.send_event_bridge_event(bus, data, name, type)
        tasks.LambdaInvoke(self.environment.scope, )
        start = sfn.Pass(self.environment.scope, 'startSelectionProcessing', input_path='$', output_path='$')
        event = sfn.Activity(self.environment.scope, 'test')
        selection_processing = sfn.StateMachine(self.environment.scope, 'selection-workflow', definition=start)
        start.next()
    def send_event_bridge_event(
            self,
            bus: events.IEventBus,
            name: str,
            type: str,
            data: list[TaskInput]
    ):
        return tasks.EventBridgePutEvents(self.environment.scope, name, entries=[
            tasks.EventBridgePutEventsEntry(
                detail_type=type,
                detail=x,
                event_bus=bus
            )
            for x in data
        ])



from aws_cdk.aws_ecs import Cluster, FargateService, IService, ICluster
from aws_cdk.aws_lambda import Function
from aws_cdk.aws_rds import DatabaseInstance, DatabaseCluster, IDatabaseCluster, IDatabaseInstance
from aws_cdk.aws_elasticsearch import IDomain, Domain
from aws_cdk.aws_elasticache import CfnCacheCluster
from aws_cdk.aws_apigateway import RestApi, IRestApi

app_services = [
    Function,
    Cluster, FargateService, IService, ICluster
]

db_services = [
    DatabaseInstance, DatabaseCluster, IDatabaseCluster, IDatabaseInstance,
    IDomain, Domain,
    CfnCacheCluster
]

web_services = [
    RestApi, IRestApi
]


from aws_cdk import aws_stepfunctions_tasks as tasks
from aws_cdk import aws_stepfunctions as sfn
from aws_cdk import aws_events as events
from aws_cdk.aws_stepfunctions import TaskInput

from infraflow.cdk.core.service_stage import ServiceStageStack


class StepFunctionWorkflow:
    def __init__(self, environment: ServiceStageStack, name: str):
        self.name = name
        self.environment = environment

        tasks.LambdaInvoke(self.environment, )
        start = sfn.Pass(self.environment, 'startSelectionProcessing', input_path='$', output_path='$')
        event = sfn.Activity(self.environment, 'test')
        selection_processing = sfn.StateMachine(self.environment, 'selection-workflow', definition=start)
        start.next()

    def start(self):
        start = sfn
        start = sfn.Pass(self.environment, f"start_${self.name}", input_path='$', output_path='$')

    def send_event_bridge_event(
            self,
            bus: events.IEventBus,
            name: str,
            type: str,
            data: list[TaskInput]
    ):
        return tasks.EventBridgePutEvents(self.environment.scope, name, entries=[
            tasks.EventBridgePutEventsEntry(
                detail_type=type,
                detail=x,
                event_bus=bus
            )
            for x in data
        ])



from infraflow.cdk.core.service_stage import ServiceStageStack
from infraflow.cdk.core.component_service import ComponentService
from infraflow.cdk.core.construct import HasEnv
from infraflow.cdk.core.environment import EnvConfig, Env, IEnv


from typing import Union, Optional
from aws_cdk import Environment as CdkEnv
from enum import Enum
import boto3

from aws_cdk.aws_ec2 import Vpc, CfnInternetGateway, NatProvider, NatInstanceProvider, Subnet, IVpc, ISubnet, \
    SubnetType, InterfaceVpcEndpoint, InterfaceVpcEndpointAwsService, VpcEndpointType, VpcEndpoint, GatewayVpcEndpoint
from constructs import IConstruct, Construct

from infraflow.priv_utils import only_truthy_items

DEFAULT_INTERFACE_SERVICES = [
    InterfaceVpcEndpointAwsService.SQS,
    InterfaceVpcEndpointAwsService.SNS,
    InterfaceVpcEndpointAwsService.S3,
    InterfaceVpcEndpointAwsService.SES,
    InterfaceVpcEndpointAwsService.RDS_DATA
]

class IEnv:
    @property
    def vpc(self) -> IVpc:
        raise NotImplemented()

    def vpc_subnets(self, subnet_type=None) -> list[ISubnet]:
        raise NotImplemented()

    def service_subnets(self, subnet_type: SubnetType):
        raise NotImplemented()

    def get_secret(self, key) -> str:
        raise NotImplemented()

    def get_param(self, key) -> str:
        raise NotImplemented()

    @property
    def environment_vars(self) -> dict:
        raise NotImplemented()

    def service_endpoints(self, services: list[InterfaceVpcEndpointAwsService]) -> list[InterfaceVpcEndpoint]:
        raise NotImplemented()


class EnvConfig:
    def __init__(
            self,
            env: CdkEnv,
            vpc_id: str,
            subnet_map: dict[SubnetType, list[str]] = None,
            environment_variables: dict[str, str] = {}
    ):
        self.env = env
        self.environment_variables = environment_variables
        self.vpc_id = vpc_id
        self.subnet_map = subnet_map


class Env(IEnv):
    def __init__(self, stage_name: str, scope: Construct, config: EnvConfig):
        self.stage_name = stage_name
        self.config = config
        self.scope = scope
        self.secrets_manager = boto3.client('secretsmanager')
        self.ssm = boto3.client('ssm')
        self._vpc = Vpc.from_lookup(self.scope, "VPC", vpc_id=self.config.vpc_id)

    @property
    def vpc(self) -> IVpc:
        return self._vpc

    def vpc_subnets(self, subnet_type: SubnetType=None) -> list[ISubnet]:
        return self.vpc.select_subnets(subnet_type=subnet_type).subnets

    def service_subnets(self, subnet_type: SubnetType):
        vpc_subnets = self.vpc_subnets(subnet_type)
        if not self.config.subnet_map:
            return vpc_subnets

        subnet_ids = self.config.subnet_map[subnet_type]
        return [sn for sn in vpc_subnets for sni in subnet_ids if sni == sn.subnet_id]

    def secret(self, key):
        return self.secrets_manager.get_secret_value(key)

    def param(self, key):
        return self.ssm.get_parameter(key)

    def service_endpoints(self, services: list[InterfaceVpcEndpointAwsService]) -> list[InterfaceVpcEndpoint]:
        return self.interface_endpoints(service_names=[s.name for s in services])

    def _get_endpoints(
            self,
            endpoint_type: VpcEndpointType, *,
            service_names: Optional[list[str]] = None,
    ) -> list[dict]:
        client = boto3.client('ec2')
        results = client.describe_vpc_endpoints(Filters=only_truthy_items([
            {
                "Name": "vpc-id",
                "Values": [self.vpc.vpc_id],
            },
            {
                "Name": "vpc-endpoint-type",
                "Values": [str(endpoint_type).lower().capitalize()],
            },
            {
                "Name": "service-name",
                "Values": service_names
            } if service_names else None
        ]))
        return results

    def interface_endpoints(self, service_names: Optional[list[str]] = None) -> list[InterfaceVpcEndpoint]:
        return [
            InterfaceVpcEndpoint.from_interface_vpc_endpoint_attributes(
                self.scope, '', vpc_endpoint_id=ep.get('VpcEndpointId')
            )
            for ep in self._get_endpoints(endpoint_type=VpcEndpointType.INTERFACE, service_names=service_names)
            if ep.get("VpcEndpointType") == "Interface"
        ]

    def gateway_endpoints(self) -> list[GatewayVpcEndpoint]:
        return [
            GatewayVpcEndpoint.from_gateway_vpc_endpoint_id(
                self.scope, '', vpc_endpoint_id=ep.get('VpcEndpointId')
            )
            for ep in self._get_endpoints(endpoint_type=VpcEndpointType.GATEWAY)
            if ep.get("VpcEndpointType") == "Gateway"
        ]


    @property
    def environment_vars(self):
        return self.config.environment_variables
